import pytest
import asyncio

from starkware.starknet.testing.starknet import Starknet
from starkware.starknet.definitions.error_codes import StarknetErrorCode
from starkware.starknet.compiler.compile import get_selector_from_name

from utils.Signer import Signer
from utils.misc import deploy, declare, deploy_proxy, assert_revert, str_to_felt, assert_event_emmited
from utils.TransactionSender import TransactionSender

signer = Signer(123456789987654321)
guardian = Signer(456789987654321123)
wrong_signer = Signer(666666666666666666)
wrong_guardian = Signer(6767676767)

VERSION = str_to_felt('0.1.0')


@pytest.fixture(scope='module')
def event_loop():
  return asyncio.new_event_loop()


@pytest.fixture(scope='module')
async def get_starknet():
  starknet = await Starknet.empty()
  return starknet


@pytest.fixture
async def account_factory(get_starknet):
  starknet = get_starknet
  account = await deploy(starknet, "contracts/account/Account.cairo")
  await account.initialize(signer.public_key, guardian.public_key).invoke()
  return account


@pytest.fixture
async def account_factory(get_starknet):
  starknet = get_starknet
  implementation_class = await declare(starknet, "contracts/account/Account.cairo")
  proxy, account = await deploy_proxy(
    starknet,
    "contracts/proxy/Proxy.cairo",
    implementation_class.abi,
    [implementation_class.class_hash, get_selector_from_name('initialize'), 2, signer.public_key, guardian.public_key],
  )
  return account, proxy, implementation_class.class_hash


@pytest.fixture
async def dapp_factory(get_starknet):
    starknet = get_starknet
    dapp_class = await declare(starknet, "contracts/test/dapp.cairo")
    dapp = await deploy(starknet, "contracts/test/dapp.cairo")
    return dapp, dapp_class.class_hash


@pytest.mark.asyncio
async def test_initializer(account_factory):
  account, proxy, implementation = account_factory
  assert (await proxy.get_implementation().call()).result.implementation == implementation
  assert (await account.get_signer_public_key().call()).result.public_key == signer.public_key
  assert (await account.get_guardian_public_key().call()).result.public_key == guardian.public_key
  assert (await account.get_version().call()).result.version == VERSION


@pytest.mark.asyncio
async def test_call_dapp(account_factory, dapp_factory):
  account, _, _ = account_factory
  dapp, _ = dapp_factory
  sender = TransactionSender(account)

  # should revert with the wrong signer
  await assert_revert(
    sender.send_transaction([(dapp.contract_address, 'set_number', [47])], wrong_signer),
    "Account: invalid signer signature"
  )

  # should call the dapp
  assert (await dapp.get_number(account.contract_address).call()).result.number == 0
  await sender.send_transaction([(dapp.contract_address, 'set_number', [47])], signer)
  assert (await dapp.get_number(account.contract_address).call()).result.number == 47

@pytest.mark.asyncio
async def test_upgrade(account_factory, dapp_factory):
  account, proxy, account_impl_1 = account_factory
  _, _, account_impl_2 = account_factory
  dapp, dapp_class_hash = dapp_factory

  sender = TransactionSender(account)

  # should revert with the wrong guardian
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'upgrade', [account_impl_2])], wrong_signer),
    "Account: invalid signer signature"
  )

  # should revert when the target is not an account
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'upgrade', [dapp_class_hash])], signer),
    "Account: invalid implementation",
    StarknetErrorCode.ENTRY_POINT_NOT_FOUND_IN_CONTRACT
  )

  assert (await proxy.get_implementation().call()).result.implementation == (account_impl_1)

  tx_exec_info = await sender.send_transaction([(account.contract_address, 'upgrade', [account_impl_2])], signer)

  assert_event_emmited(
      tx_exec_info,
      from_address=account.contract_address,
      name='AccountUpgraded',
      data=[account_impl_2]
  )

  assert (await proxy.get_implementation().call()).result.implementation == (account_impl_2)



from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.definitions.general_config import StarknetChainId
from starkware.starknet.core.os.transaction_hash.transaction_hash import calculate_transaction_hash_common, TransactionHashPrefix
import logging

LOGGER = logging.getLogger(__name__)

TRANSACTION_VERSION = 0

class TransactionSender():
  def __init__(self, account):
    self.account = account

  async def send_transaction(self, calls, signer, nonce=None, max_fee=0):
    if nonce is None:
      execution_info = await self.account.get_nonce().call()
      nonce = execution_info.result.nonce

    calls_with_selector = [(call[0], get_selector_from_name(call[1]), call[2]) for call in calls]
    call_array, calldata = from_call_to_call_array(calls)

    transaction_hash = get_transaction_hash(self.account.contract_address, call_array, calldata, nonce, max_fee)

    signature = list(signer.sign(transaction_hash))

    return await self.account.__execute__(call_array, calldata, nonce).invoke(signature=signature)

def from_call_to_call_array(calls):
  call_array = []
  calldata = []
  for call in calls:
    assert len(call) == 3, "Invalid call parameters"
    entry = (call[0], get_selector_from_name(call[1]), len(calldata), len(call[2]))
    call_array.append(entry)
    calldata.extend(call[2])
  return call_array, calldata

def get_transaction_hash(account, call_array, calldata, nonce, max_fee):
  execute_calldata = [
    len(call_array),
    *[x for t in call_array for x in t],
    len(calldata),
    *calldata,
    nonce]

  return calculate_transaction_hash_common(
    TransactionHashPrefix.INVOKE,
    TRANSACTION_VERSION,
    account,
    get_selector_from_name('__execute__'),
    execute_calldata,
    max_fee,
    StarknetChainId.TESTNET.value,
    []
  )



import pytest
import asyncio
import logging

from starkware.starknet.testing.starknet import Starknet
from starkware.starknet.business_logic.state.state import BlockInfo

from utils.Signer import Signer
from utils.misc import deploy, assert_revert, str_to_felt, assert_event_emmited
from utils.TransactionSender import TransactionSender

LOGGER = logging.getLogger(__name__)

signer = Signer(123456789987654321)
guardian = Signer(456789987654321123)

wrong_signer = Signer(666666666666666666)
wrong_guardian = Signer(6767676767)

DEFAULT_TIMESTAMP = 1640991600
ESCAPE_SECURITY_PERIOD = 24 * 7 * 60 * 60

VERSION = str_to_felt('0.1.0')

IACCOUNT_ID = 0xf10dbd44

@pytest.fixture(scope='module')
def event_loop():
  return asyncio.new_event_loop()

@pytest.fixture(scope='module')
async def get_starknet():
  starknet = await Starknet.empty()
  return starknet

def update_starknet_block(starknet, block_number=1, block_timestamp=DEFAULT_TIMESTAMP):
  starknet.state.state.block_info = BlockInfo(
    block_number=block_number,
    block_timestamp=block_timestamp,
    gas_price=0,
    sequencer_address=starknet.state.state.block_info.sequencer_address
  )

def reset_starknet_block(starknet):
  update_starknet_block(starknet=starknet)


@pytest.fixture
async def account_factory(get_starknet):
  starknet = get_starknet
  account = await deploy(starknet, "contracts/account/Account.cairo")
  await account.initialize(signer.public_key, guardian.public_key).invoke()
  return account


@pytest.fixture
async def dapp_factory(get_starknet):
  starknet = get_starknet
  dapp = await deploy(starknet, "contracts/test/dapp.cairo")
  return dapp


@pytest.mark.asyncio
async def test_initializer(account_factory):
  account = account_factory
  # should be configured correctly
  assert (await account.get_signer_public_key().call()).result.public_key == (signer.public_key)
  assert (await account.get_guardian_public_key().call()).result.public_key == (guardian.public_key)
  assert (await account.get_version().call()).result.version == VERSION
  assert (await account.supportsInterface(IACCOUNT_ID).call()).result.success == 1
  # should throw when calling initialize twice
  await assert_revert(
    account.initialize(signer.public_key, guardian.public_key).invoke(),
    "Account: already initialized"
  )


@pytest.mark.asyncio
async def test_call_dapp(account_factory, dapp_factory):
  account = account_factory
  dapp = dapp_factory
  sender = TransactionSender(account)

  calls = [(dapp.contract_address, 'set_number', [47])]

  # should revert with the wrong nonce
  await assert_revert(
    sender.send_transaction(calls, signer, nonce=3),
    "Account: invalid nonce"
  )

  # should revert with the wrong signer
  await assert_revert(
    sender.send_transaction(calls, wrong_signer),
    "Account: invalid signer signature"
  )

  # should call the dapp
  assert (await dapp.get_number(account.contract_address).call()).result.number == 0

  tx_exec_info = await sender.send_transaction(calls, signer)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='TransactionExecuted'
  )

  assert (await dapp.get_number(account.contract_address).call()).result.number == 47


@pytest.mark.asyncio
async def test_multicall(account_factory, dapp_factory):
  account = account_factory
  dapp = dapp_factory
  sender = TransactionSender(account)

  # should reverts when one of the call is to the account
  await assert_revert(
    sender.send_transaction([
      (dapp.contract_address, 'set_number', [47]),
      (account.contract_address, 'set_signer_public_key', [1])
    ], signer)
  )

  await assert_revert(
    sender.send_transaction([
      (account.contract_address, 'set_signer_public_key', [1]),
      (dapp.contract_address, 'set_number', [47])
    ], signer)
  )

  # should call the dapp
  assert (await dapp.get_number(account.contract_address).call()).result.number == 0

  await sender.send_transaction([
    (dapp.contract_address, 'set_number', [47]),
    (dapp.contract_address, 'increase_number', [10])
  ], signer)

  assert (await dapp.get_number(account.contract_address).call()).result.number == 57


@pytest.mark.asyncio
async def test_change_signer(account_factory):
  account = account_factory
  sender = TransactionSender(account)
  new_signer = Signer(4444444444)

  assert (await account.get_signer_public_key().call()).result.public_key == (signer.public_key)

  # should revert with the wrong signer
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'set_signer_public_key', [new_signer.public_key])], wrong_signer),
    "Account: invalid signer signature"
  )

  # should work with the correct signers
  tx_exec_info = await sender.send_transaction([(account.contract_address, 'set_signer_public_key', [new_signer.public_key])], signer)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='SignerPublicKeyChanged',
    data=[new_signer.public_key]
  )

  assert (await account.get_signer_public_key().call()).result.public_key == (new_signer.public_key)


@pytest.mark.asyncio
async def test_change_guardian(account_factory):
  account = account_factory
  sender = TransactionSender(account)
  new_guardian = Signer(55555555)

  assert (await account.get_guardian_public_key().call()).result.public_key == (guardian.public_key)

  # should revert with the wrong signer
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'set_guardian_public_key', [new_guardian.public_key])], wrong_signer),
    "Account: invalid signer signature"
  )

  # should work with the correct signers
  tx_exec_info = await sender.send_transaction([(account.contract_address, 'set_guardian_public_key', [new_guardian.public_key])], signer)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='GuardianPublicKeyChanged',
    data=[new_guardian.public_key]
  )

  assert (await account.get_guardian_public_key().call()).result.public_key == (new_guardian.public_key)


@pytest.mark.asyncio
async def test_trigger_escape_signer_by_guardian(get_starknet, account_factory):
  account = account_factory
  starknet = get_starknet
  sender = TransactionSender(account)

  # reset block_timestamp
  reset_starknet_block(starknet=starknet)

  escape = (await account.get_signer_escape().call()).result
  assert (escape.active_at == 0)

  # should revert with the wrong guardian
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], wrong_guardian),
    "Account: invalid guardian signature"
  )

  tx_exec_info = await sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], guardian)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='SignerEscapeTriggered',
    data=[DEFAULT_TIMESTAMP + ESCAPE_SECURITY_PERIOD]
  )

  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == (DEFAULT_TIMESTAMP + ESCAPE_SECURITY_PERIOD)


@pytest.mark.asyncio
async def test_escape_signer(get_starknet, account_factory):
  account = account_factory
  starknet = get_starknet
  sender = TransactionSender(account)
  new_signer = Signer(5555555578895)

  # reset block_timestamp
  reset_starknet_block(starknet=starknet)

  # trigger escape
  await sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], guardian)
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == (DEFAULT_TIMESTAMP + ESCAPE_SECURITY_PERIOD)

  # should fail to escape before the end of the period
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'escape_signer', [new_signer.public_key])], guardian),
    "Account: invalid escape"
  )

  # should revert with the wrong guardian
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'escape_signer', [new_signer.public_key])], wrong_guardian),
    "Account: invalid guardian signature"
  )

  # wait security period
  update_starknet_block(starknet=starknet, block_timestamp=(DEFAULT_TIMESTAMP+ESCAPE_SECURITY_PERIOD))

  # should escape after the security period
  assert (await account.get_signer_public_key().call()).result.public_key == (signer.public_key)
  tx_exec_info = await sender.send_transaction([(account.contract_address, 'escape_signer', [new_signer.public_key])], guardian)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='SignerEscaped',
    data=[new_signer.public_key]
  )

  assert (await account.get_signer_public_key().call()).result.public_key == (new_signer.public_key)

  # escape should be cleared
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == 0


@pytest.mark.asyncio
async def test_guardian_overrides_trigger_escape_signer(get_starknet, account_factory):
  account = account_factory
  starknet = get_starknet
  sender = TransactionSender(account)
  new_signer = Signer(5555555578895)

  # reset block_timestamp
  reset_starknet_block(starknet=starknet)

  # trigger escape
  await sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], guardian)
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == (DEFAULT_TIMESTAMP + ESCAPE_SECURITY_PERIOD)

  # wait few seconds
  update_starknet_block(starknet=starknet, block_timestamp=(DEFAULT_TIMESTAMP+100))

  # signer overrides escape
  await sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], guardian)
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == (DEFAULT_TIMESTAMP + 100 + ESCAPE_SECURITY_PERIOD)


@pytest.mark.asyncio
async def test_cancel_escape(get_starknet, account_factory):
  account = account_factory
  starknet = get_starknet
  sender = TransactionSender(account)

  # reset block_timestamp
  reset_starknet_block(starknet=starknet)

  # trigger escape
  await sender.send_transaction([(account.contract_address, 'trigger_signer_escape', [])], guardian)
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == (DEFAULT_TIMESTAMP + ESCAPE_SECURITY_PERIOD)

  # should fail to cancel with the guardian
  await assert_revert(
    sender.send_transaction([(account.contract_address, 'cancel_escape', [])], guardian),
    "Account: invalid signer signature"
  )

  # cancel escape
  tx_exec_info = await sender.send_transaction([(account.contract_address, 'cancel_escape', [])], signer)

  assert_event_emmited(
    tx_exec_info,
    from_address=account.contract_address,
    name='EscapeCanceled',
    data=[]
  )

  # escape should be cleared
  escape = (await account.get_signer_escape().call()).result
  assert escape.active_at == 0


@pytest.mark.asyncio
async def test_is_valid_signature(account_factory):
  account = account_factory
  hash = 1283225199545181604979924458180358646374088657288769423115053097913173815464

  signature = list(signer.sign(hash))

  res = (await account.is_valid_signature(hash, signature).call()).result
  assert (res.is_valid == 1)



from setuptools import setup

def local_scheme(version):
  """Skip the local version (eg. +xyz of 0.6.1.dev4+gdf99fe2)
  to be able to upload to Test PyPI"""
  return ""

if __name__ == "__main__":
  try:
    setup(use_scm_version={'local_scheme': local_scheme})
  except:  # noqa
    print(
      "\n\nAn error occurred while building the project, "
      "please ensure you have the most updated version of setuptools, "
      "setuptools_scm and wheel with:\n"
      "   pip install -U setuptools setuptools_scm wheel\n\n"
    )
    raise
